@@ -42,8 +42,8 @@
 #include <math.h>
 #include <stdio.h>
 
-static void     eel_background_class_init          (gpointer       klass);
-static void     eel_background_init                (gpointer       object,
+static void     eel_background_class_init                (gpointer       klass);
+static void     eel_background_init                      (gpointer       object,
 							  gpointer       klass);
 static void     eel_background_finalize                  (GObject       *object);
 static void     eel_background_start_loading_image       (EelBackground *background,
@@ -61,7 +61,10 @@ enum {
 	LAST_SIGNAL
 };
 
-#define	RESET_BACKGROUND_MAGIC_IMAGE_NAME "reset.png"
+/* This is the size of the GdkRGB dither matrix, in order to avoid
+ * bad dithering when tiling the gradient
+ */
+#define GRADIENT_PIXMAP_TILE_SIZE 128
 
 static guint signals[LAST_SIGNAL];
 
@@ -74,6 +77,8 @@ struct EelBackgroundDetails {
 
 	gboolean is_solid_color;
 	GdkColor solid_color;
+
+	gboolean constant_size;
 	
 	char *image_uri;
 	GdkPixbuf *image;
@@ -92,6 +97,13 @@ struct EelBackgroundDetails {
 	int image_rect_y;
 	int image_rect_width;
 	int image_rect_height;
+
+	/* Realized data: */
+	GdkPixmap *background_pixmap;
+	int background_entire_width;
+	int background_entire_height;
+	GdkColor background_color;
+	gboolean background_changes_with_size;
 };
 
 static void
@@ -153,9 +165,10 @@ eel_background_init (gpointer object, gpointer klass)
 {
 	EelBackground *background;
 
-	background = EEL_BACKGROUND(object);
+	background = EEL_BACKGROUND (object);
 
 	background->details = g_new0 (EelBackgroundDetails, 1);
+	background->details->constant_size = FALSE;
 }
 
 /* The safe way to clear an image from a background is:
@@ -186,11 +199,26 @@ eel_background_finalize (GObject *object)
 	g_free (background->details->gradient_buffer);
 	g_free (background->details->image_uri);
 	eel_background_remove_current_image (background);
+
+	if (background->details->background_pixmap != NULL) {
+		g_object_unref (background->details->background_pixmap);
+		background->details->background_pixmap = NULL;
+	}
+
 	g_free (background->details);
 
 	EEL_CALL_PARENT (G_OBJECT_CLASS, finalize, (object));
 }
 
+void
+eel_background_set_is_constant_size (EelBackground *background,
+				     gboolean       constant_size)
+{
+	g_return_if_fail (EEL_IS_BACKGROUND (background));
+	
+	background->details->constant_size = constant_size;
+}
+
 EelBackgroundImagePlacement
 eel_background_get_image_placement (EelBackground *background)
 {
@@ -201,7 +229,7 @@ eel_background_get_image_placement (EelBackground *background)
 
 static gboolean
 eel_background_set_image_placement_no_emit (EelBackground *background,
-						 EelBackgroundImagePlacement new_placement)
+					    EelBackgroundImagePlacement new_placement)
 {
 	g_return_val_if_fail (EEL_IS_BACKGROUND (background), FALSE);
 
@@ -453,7 +481,11 @@ eel_background_image_totally_obscures (EelBackground *background)
 static void
 eel_background_ensure_image_scaled (EelBackground *background, int dest_width, int dest_height)
 {
-	if (background->details->image == NULL) {
+	/* Avoid very bad gdk-pixbuf scaling behaviour on 1x1 (not yet
+	 * sized) windows and large backrounds
+	 */
+	if (background->details->image == NULL ||
+	    (dest_width == 1 && dest_height == 1)) {
 		background->details->image_rect_x = 0;
 		background->details->image_rect_y = 0;
 		background->details->image_rect_width = 0;
@@ -494,9 +526,9 @@ eel_background_ensure_image_scaled (EelBackground *background, int dest_width, i
 			break;
 		case EEL_BACKGROUND_SCALED_ASPECT:
 			eel_gdk_scale_to_fit_factor (background->details->image_width_unscaled,
-							  background->details->image_height_unscaled,
-							  dest_width, dest_height,
-							  &fit_width, &fit_height);
+						     background->details->image_height_unscaled,
+						     dest_width, dest_height,
+						     &fit_width, &fit_height);
 			if (image_width != fit_width || image_height != fit_height) {
 				if (cur_scaled) {
 					reload_image = TRUE;
@@ -533,6 +565,138 @@ eel_background_ensure_image_scaled (EelBackground *background, int dest_width, i
 	}
 }
 
+static gboolean
+get_pixmap_size (EelBackground    *background,
+		 int               entire_width,
+		 int               entire_height,
+		 int              *pixmap_width,
+		 int              *pixmap_height,
+		 gboolean         *changes_with_size)
+{
+	*pixmap_width = 0;	
+	*pixmap_height = 0;
+	*changes_with_size = ! background->details->constant_size;
+	
+	if (background->details->image == NULL) {
+		if (background->details->is_solid_color) {
+			*changes_with_size = FALSE;
+			return FALSE;
+		}
+		if (background->details->gradient_is_horizontal) {
+			*pixmap_width = entire_width;
+			*pixmap_height = GRADIENT_PIXMAP_TILE_SIZE;
+		} else {
+			*pixmap_width = GRADIENT_PIXMAP_TILE_SIZE;
+			*pixmap_height = entire_height;
+		}
+		return TRUE;
+	} else if (!eel_background_is_image_load_in_progress (background)) {
+		switch (background->details->image_placement) {
+		case EEL_BACKGROUND_TILED:
+			if (background->details->is_solid_color || !gdk_pixbuf_get_has_alpha (background->details->image)) {
+				*pixmap_width = background->details->image_width_unscaled;
+				*pixmap_height = background->details->image_height_unscaled;
+				*changes_with_size = FALSE;
+			} else {
+				if (background->details->gradient_is_horizontal) {
+					*pixmap_width = entire_width;
+					*pixmap_height = background->details->image_height_unscaled;
+				} else {
+					*pixmap_width = background->details->image_width_unscaled;
+					*pixmap_height = entire_height;
+				}
+			}
+			break;
+		case EEL_BACKGROUND_CENTERED:
+		case EEL_BACKGROUND_SCALED:
+		case EEL_BACKGROUND_SCALED_ASPECT:
+			*pixmap_width = entire_width;
+			*pixmap_height = entire_height;
+		}
+		return TRUE;
+	}
+	return FALSE;
+}
+
+
+static void
+eel_background_unrealize (EelBackground *background)
+{
+	if (background->details->background_pixmap != NULL) {
+		g_object_unref (background->details->background_pixmap);
+		background->details->background_pixmap = NULL;
+	}
+	background->details->background_entire_width = 0;
+	background->details->background_entire_height = 0;
+}
+
+static void
+eel_background_ensure_realized (EelBackground *background, GdkWindow *window,
+				int entire_width, int entire_height)
+{
+	int pixmap_width, pixmap_height;
+	char *start_color_spec;
+ 	GdkPixmap *pixmap;
+	GdkGC *gc;
+
+	start_color_spec = eel_gradient_get_start_color_spec (background->details->color);
+	eel_gdk_color_parse_with_white_default (start_color_spec, &background->details->background_color);
+	g_free (start_color_spec);
+
+	/* If the pixmap doesn't change with the window size, never update
+	 * it again.
+	 */
+	if (background->details->background_pixmap != NULL &&
+	    !background->details->background_changes_with_size) {
+		return;
+	}
+
+	/* If the window size is the same as last time, don't update */
+	if (entire_width == background->details->background_entire_width &&
+	    entire_height == background->details->background_entire_height) {
+		return;
+	}
+
+	if (background->details->background_pixmap != NULL) {
+		g_object_unref (background->details->background_pixmap);
+		background->details->background_pixmap = NULL;
+	}
+	
+	if (get_pixmap_size (background, entire_width, entire_height,
+			      &pixmap_width, &pixmap_height, &background->details->background_changes_with_size)) {
+		pixmap = gdk_pixmap_new (window, pixmap_width, pixmap_height, -1);
+		gc = gdk_gc_new (pixmap);
+		eel_background_pre_draw (background,  entire_width, entire_height);
+		eel_background_draw (background, pixmap, gc,
+				     0, 0, 0, 0,
+				     pixmap_width, pixmap_height);
+		g_object_unref (gc);
+		background->details->background_pixmap = pixmap;
+	}
+
+	background->details->background_entire_width = entire_width;
+	background->details->background_entire_height = entire_height;
+}
+
+GdkPixmap *
+eel_background_get_pixmap_and_color (EelBackground *background,
+				     GdkWindow     *window,
+				     int            entire_width,
+				     int            entire_height,
+				     GdkColor      *color,
+				     gboolean      *changes_with_size)
+{
+	eel_background_ensure_realized (background, window, entire_width, entire_height);
+	
+	*color = background->details->background_color;
+	*changes_with_size = background->details->background_changes_with_size;
+	
+	if (background->details->background_pixmap != NULL) {
+		return g_object_ref (background->details->background_pixmap);
+	} 
+	return NULL;
+}
+
 void
 eel_background_pre_draw (EelBackground *background, int entire_width, int entire_height)
 {
@@ -592,10 +756,10 @@ eel_background_draw (EelBackground *background,
 
 void
 eel_background_draw_to_drawable (EelBackground *background,
-				      GdkDrawable *drawable, GdkGC *gc,
-				      int drawable_x, int drawable_y,
-				      int drawable_width, int drawable_height,
-				      int entire_width, int entire_height)
+				 GdkDrawable *drawable, GdkGC *gc,
+				 int drawable_x, int drawable_y,
+				 int drawable_width, int drawable_height,
+				 int entire_width, int entire_height)
 {
 	eel_background_pre_draw (background, entire_width, entire_height);
 	eel_background_draw (background, drawable, gc,
@@ -606,13 +770,13 @@ eel_background_draw_to_drawable (EelBackground *background,
 
 void
 eel_background_draw_to_pixbuf (EelBackground *background,
-				    GdkPixbuf *pixbuf,
-				    int pixbuf_x,
-				    int pixbuf_y,
-				    int pixbuf_width,
-				    int pixbuf_height,
-				    int entire_width,
-				    int entire_height)
+			       GdkPixbuf *pixbuf,
+			       int pixbuf_x,
+			       int pixbuf_y,
+			       int pixbuf_width,
+			       int pixbuf_height,
+			       int entire_width,
+			       int entire_height)
 {
 	GnomeCanvasBuf fake_buffer;
 
@@ -622,9 +786,9 @@ eel_background_draw_to_pixbuf (EelBackground *background,
 	canvas_buf_from_pixbuf (&fake_buffer, pixbuf, pixbuf_x, pixbuf_y, pixbuf_width, pixbuf_height);
 
 	eel_background_draw_to_canvas (background,
-					    &fake_buffer,
-					    entire_width,
-					    entire_height);
+				       &fake_buffer,
+				       entire_width,
+				       entire_height);
 }
 
 /* fill the canvas buffer with a tiled pixbuf */
@@ -705,9 +869,9 @@ eel_background_draw_aa (EelBackground *background, GnomeCanvasBuf *buffer)
 
 void
 eel_background_draw_to_canvas (EelBackground *background,
-				    GnomeCanvasBuf *buffer,
-				    int entire_width,
-				    int entire_height)
+			       GnomeCanvasBuf *buffer,
+			       int entire_width,
+			       int entire_height)
 {
 	eel_background_pre_draw (background, entire_width, entire_height);
 	eel_background_draw_aa (background, buffer);
@@ -731,7 +895,7 @@ eel_background_get_image_uri (EelBackground *background)
 
 static gboolean
 eel_background_set_color_no_emit (EelBackground *background,
-			       const char *color)
+				  const char *color)
 {
 	g_return_val_if_fail (EEL_IS_BACKGROUND (background), FALSE);
 
@@ -747,7 +911,7 @@ eel_background_set_color_no_emit (EelBackground *background,
 
 void
 eel_background_set_color (EelBackground *background,
-			       const char *color)
+			  const char *color)
 {
 	if (eel_background_set_color_no_emit (background, color)) {
 		g_signal_emit (G_OBJECT (background), signals[SETTINGS_CHANGED], 0);
@@ -759,8 +923,8 @@ eel_background_set_color (EelBackground *background,
 
 static void
 eel_background_load_image_callback (GnomeVFSResult error,
-					 GdkPixbuf *pixbuf,
-					 gpointer callback_data)
+				    GdkPixbuf *pixbuf,
+				    gpointer callback_data)
 {
 	EelBackground *background;
 
@@ -806,6 +970,7 @@ static void
 eel_background_start_loading_image (EelBackground *background, gboolean emit_appearance_change,
 				    gboolean load_async)
 {
+	GdkPixbuf *pixbuf;
 	background->details->emit_after_load = emit_appearance_change;
 
 	if (background->details->image_uri != NULL) {
@@ -816,7 +981,13 @@ eel_background_start_loading_image (EelBackground *background, gboolean emit_app
 											    background);
 		}
 		else {
-			eel_background_load_image_callback (0, eel_gdk_pixbuf_load (background->details->image_uri), background);
+			pixbuf = eel_gdk_pixbuf_load (background->details->image_uri);
+			eel_background_load_image_callback (0, pixbuf, background);
+
+			/* load_image_callback refs the pixbuf so we unref it here */
+			if (pixbuf != NULL) {
+				g_object_unref (pixbuf);
+			}
 		}
 	} else {
 		eel_background_load_image_callback (0, NULL, background);
@@ -825,9 +996,9 @@ eel_background_start_loading_image (EelBackground *background, gboolean emit_app
 
 static gboolean
 eel_background_set_image_uri_helper (EelBackground *background,
-					const char *image_uri,
-					gboolean emit_setting_change,
-					gboolean emit_appearance_change,
+				     const char *image_uri,
+				     gboolean emit_setting_change,
+				     gboolean emit_appearance_change,
 				     gboolean load_async)
 {
 	g_return_val_if_fail (EEL_IS_BACKGROUND (background), FALSE);
@@ -870,10 +1041,9 @@ eel_background_set_image_uri_sync (EelBackground *background, const char *image_
 static void
 set_image_and_color_image_loading_done_callback (EelBackground *background, gboolean successful_load, char *color)
 {
-	g_signal_handlers_disconnect_by_func (
-		G_OBJECT (background),
-		G_CALLBACK (set_image_and_color_image_loading_done_callback),
-		(gpointer) color);
+	g_signal_handlers_disconnect_by_func
+		(G_OBJECT (background),
+		 G_CALLBACK (set_image_and_color_image_loading_done_callback), color);
 
 	eel_background_set_color_no_emit (background, color);
 
@@ -902,10 +1072,8 @@ eel_background_set_image_uri_and_color (EelBackground *background, const char *i
 
 	color_copy = g_strdup (color);
 
-	g_signal_connect (background,
-			    "image_loading_done",
-			    G_CALLBACK (set_image_and_color_image_loading_done_callback),
-			    (gpointer) color_copy);
+	g_signal_connect (background, "image_loading_done",
+			  G_CALLBACK (set_image_and_color_image_loading_done_callback), color_copy);
 			    
 	/* set_image_and_color_image_loading_done_callback must always be called
 	 * because we rely on it to:
@@ -926,22 +1094,40 @@ eel_background_set_image_uri_and_color (EelBackground *background, const char *i
 
 void
 eel_background_receive_dropped_background_image (EelBackground *background,
-						      const char *image_uri)
+						 const char *image_uri)
 {
-	/* Special case the reset-background image by file name. */
-	if (eel_str_has_suffix (image_uri, "/" RESET_BACKGROUND_MAGIC_IMAGE_NAME)) {
-		eel_background_reset (background);
-	} else {
-		/* Currently, we only support tiled images. So we set the placement.
-		 * We rely on eel_background_set_image_uri_and_color to emit
-		 * the SETTINGS_CHANGED & APPEARANCE_CHANGE signals.
-		 */
-		eel_background_set_image_placement_no_emit (background, EEL_BACKGROUND_TILED);
+	/* Currently, we only support tiled images. So we set the placement.
+	 * We rely on eel_background_set_image_uri_and_color to emit
+	 * the SETTINGS_CHANGED & APPEARANCE_CHANGE signals.
+	 */
+	eel_background_set_image_placement_no_emit (background, EEL_BACKGROUND_TILED);
+	
+	eel_background_set_image_uri_and_color (background, image_uri, NULL);
+}
 
-		eel_background_set_image_uri_and_color (background, image_uri, NULL);
+static void
+eel_background_set_style_font_from_default (GtkStyle *style,
+					    GtkSettings *settings)
+{
+	char *fontname;
+
+	fontname = NULL;
+	
+	g_object_get (settings,
+		      "gtk-font-name", &fontname,
+		      NULL);
+
+	if (fontname != NULL) {
+		if (style->font_desc) {
+			pango_font_description_free (style->font_desc);
+		}
+		
+		style->font_desc = pango_font_description_from_string (fontname);
+		g_free (fontname);
 	}
 }
 
+
 static void
 eel_background_set_widget_style (EelBackground *background,
 				 GtkWidget *widget)
@@ -952,7 +1138,11 @@ eel_background_set_widget_style (EelBackground *background,
 	g_return_if_fail (GTK_IS_WIDGET (widget));
 	
 	style = GTK_STYLE (eel_background_style_new (gtk_widget_get_style (widget),
-						     background));	
+						     background));
+	
+	/* Bad hack to make font changing work: */
+	eel_background_set_style_font_from_default (style, gtk_widget_get_settings (widget));
+	
 	gtk_widget_set_style (widget, style);
 	g_object_unref (style);
 }
@@ -1002,6 +1192,8 @@ draw_background_callback (GnomeCanvas *canvas, GdkDrawable *drawable,
 			  int x, int y, int width, int height)
 {
 	EelBackground *background;
+	GdkGCValuesMask value_mask;
+	GdkGCValues gc_values;
 	GdkGC *gc;
 
 	background = eel_get_widget_background (GTK_WIDGET (canvas));
@@ -1009,17 +1201,33 @@ draw_background_callback (GnomeCanvas *canvas, GdkDrawable *drawable,
 		return;
 	}
 
-	eel_background_pre_draw (background,
-				 GTK_WIDGET (canvas)->allocation.width,
-				 GTK_WIDGET (canvas)->allocation.height);
+	
+	eel_background_ensure_realized (background, GTK_WIDGET (canvas)->window,
+					MAX (GTK_LAYOUT (canvas)->hadjustment->upper, GTK_WIDGET (canvas)->allocation.width),
+					MAX (GTK_LAYOUT (canvas)->vadjustment->upper, GTK_WIDGET (canvas)->allocation.height));
+
 
 	/* Create a new gc each time.
 	 * If this is a speed problem, we can create one and keep it around,
 	 * but it's a bit more complicated to ensure that it's always compatible
 	 * with whatever drawable is passed in.
 	 */
-	gc = gdk_gc_new (drawable);
-	eel_background_draw (background, drawable, gc, x, y, 0, 0, width, height);
+	if (background->details->background_pixmap != NULL) {
+		gc_values.tile = background->details->background_pixmap;
+		gc_values.ts_x_origin = -x;
+		gc_values.ts_y_origin = -y;
+		gc_values.fill = GDK_TILED;
+		value_mask = GDK_GC_FILL | GDK_GC_TILE |
+			GDK_GC_TS_X_ORIGIN | GDK_GC_TS_Y_ORIGIN;
+	} else {
+		gc_values.foreground = background->details->background_color;
+		gdk_rgb_find_color (gtk_widget_get_colormap (GTK_WIDGET (canvas)), &gc_values.foreground);
+		gc_values.fill = GDK_SOLID;
+		value_mask = GDK_GC_FILL | GDK_GC_FOREGROUND;
+	}
+		
+    	gc = gdk_gc_new_with_values (drawable, &gc_values, value_mask);
+	gdk_draw_rectangle (drawable, gc, TRUE, 0, 0, width, height);
 	g_object_unref (gc);
 
 	/* We don't want the draw from the base class to happen. */
@@ -1069,12 +1277,32 @@ eel_background_set_up_canvas (GtkWidget *widget)
 static void
 eel_widget_background_changed (GtkWidget *widget, EelBackground *background)
 {
+	eel_background_unrealize (background);
 	eel_background_set_widget_style (background, widget);
 	eel_background_set_up_canvas (widget);
 
 	gtk_widget_queue_draw (widget);
 }
 
+/* Bad hack to make font changes work for EelBackgrounds */
+static void
+eel_widget_font_changed (GtkWidget *widget,
+			 GParamSpec *pspec,
+			 GtkSettings *settings)
+{
+	EelBackground *background;
+	GtkStyle *style;
+
+	background = eel_get_widget_background (widget);
+
+	style = gtk_style_copy (gtk_widget_get_style (widget));
+	
+	eel_background_set_style_font_from_default (style, settings);
+		
+	gtk_widget_set_style (widget, style);
+	g_object_unref (style);
+}
+
 /* Gets the background attached to a widget.
 
    If the widget doesn't already have a EelBackground object,
@@ -1090,6 +1318,7 @@ eel_get_widget_background (GtkWidget *widget)
 {
 	gpointer data;
 	EelBackground *background;
+	GtkSettings *settings;
 
 	g_return_val_if_fail (GTK_IS_WIDGET (widget), NULL);
 
@@ -1112,6 +1341,9 @@ eel_get_widget_background (GtkWidget *widget)
 				 G_CALLBACK (eel_widget_background_changed), widget, G_CONNECT_SWAPPED);
 	eel_widget_background_changed (widget, background);
 
+	settings = gtk_widget_get_settings (widget);
+	g_signal_connect_object (settings, "notify::gtk-font-name",
+				 G_CALLBACK (eel_widget_font_changed), widget, G_CONNECT_SWAPPED);
 	return background;
 }
 
@@ -1139,25 +1371,6 @@ eel_gtk_widget_find_background_ancestor (GtkWidget *widget)
 	return NULL;
 }
 
-gboolean
-eel_background_is_too_complex_for_gtk_style (EelBackground *background)
-{
-	if (background == NULL) {
-		return FALSE;
-	}
-	
-	g_return_val_if_fail (EEL_IS_BACKGROUND (background), FALSE);
-	
-	if (background->details->image_uri != NULL) {
-		return TRUE;
-	}
-	if (!background->details->is_solid_color) {
-		return TRUE;
-	}
-	
-	return FALSE;
-}
-
 /* determine if a background is darker or lighter than average, to help clients know what
    colors to draw on top with */
 gboolean
