@@ -179,4 +179,9 @@ void                 eel_gdk_pixbuf_draw_layout_clipped       (GdkPixbuf
 							       guint32                color,
 							       PangoLayout           *layout);
 
+/* Scales large pixbufs down fast */
+GdkPixbuf *          eel_gdk_pixbuf_scale_down                (GdkPixbuf *pixbuf,
+							       int dest_width,
+							       int dest_height);
+
 #endif /* EEL_GDK_PIXBUF_EXTENSIONS_H */
